@@ -15,10 +15,15 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <errno.h>
 #include <string.h>
 #include <signal.h>
 #include <unistd.h>
+#include <sys/types.h>
 #include <sys/time.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <netdb.h>
 #ifndef NO_SSL
 #include <openssl/ssl.h>
 #include "mssl.h"
@@ -29,11 +34,18 @@
 #include "io.h"
 #include "str.h"
 #include "tcp.h"
+#include "res.h"
 #include "utils.h"
 #include "error.h"
 
 static volatile int stop = 0;
 
+int quiet = 0;
+char machine_readable = 0;
+char nagios_mode = 0;
+
+char *last_error;
+
 void version(void)
 {
 	fprintf(stderr, "HTTPing v" VERSION ", (C) 2003-2005 folkert@vanheusden.com\n");
@@ -63,10 +75,31 @@ void usage(void)
 	fprintf(stderr, "		default is 200, use with -e\n");
 	fprintf(stderr, "-e str		string to display when http result code\n");
 	fprintf(stderr, "		doesn't match\n");
+	fprintf(stderr, "-I str		use 'str' for the UserAgent header\n");
+	fprintf(stderr, "-R str		use 'str' for the Referer header\n");
+	fprintf(stderr, "-r		resolve hostname only once (usefull when\n");
+	fprintf(stderr, "		pinging roundrobin DNS: also takes the first\n");
+	fprintf(stderr, "		DNS lookup out of the loop so that the first\n");
+	fprintf(stderr, "		measurement is also correct)\n");
+	fprintf(stderr, "-n warn,crit	Nagios-mode: return 1 when avg. response time\n");
+	fprintf(stderr, "		>= warn, 2 if >= crit, otherwhise return 0\n");
+	fprintf(stderr, "-N x		Nagios mode 2: return 0 when all fine, 'x'\n");
+	fprintf(stderr, "		when anything failes\n");
 	fprintf(stderr, "-q		quiet, only returncode\n");
 	fprintf(stderr, "-V		show the version\n\n");
 }
 
+void emit_error()
+{
+	if (!quiet && !machine_readable && !nagios_mode)
+	{
+		printf("%s", last_error);
+	}
+
+	if (!nagios_mode)
+		last_error[0] = 0x00;
+}
+
 void handler(int sig)
 {
 	stop = 1;
@@ -82,18 +115,27 @@ int main(int argc, char *argv[])
 	int req_len;
 	int c;
 	int count = -1, curncount = 0;
-	int wait = 1, quiet = 0;
+	int wait = 1;
 	int ok = 0, err = 0;
 	double min = 999999999999999.0, avg = 0.0, max = 0.0;
 	int timeout=30;
 	char show_statuscodes = 0;
 	char use_ssl = 0;
 	SSL_CTX *client_ctx = NULL;
-	char machine_readable = 0;
 	char *ok_str = "200";
 	char *err_str = "-1";
+	char *useragent = NULL;
+	char *referer = NULL;
+	char *host;
+	int port;
+	struct sockaddr_in addr;
+	char resolve_once = 0;
+	char have_resolved = 0;
+	double nagios_warn=0.0, nagios_crit=0.0;
+	int nagios_exit_code = 2;
+	double avg_httping_time = -1.0;
 
-        while((c = getopt(argc, argv, "g:h:p:c:i:x:t:o:e:flqsmV?")) != -1)
+        while((c = getopt(argc, argv, "g:h:p:c:i:x:t:o:e:flqsmV?I:R:rn:N:")) != -1)
         {
                 switch(c)
                 {
@@ -113,6 +155,10 @@ int main(int argc, char *argv[])
 			get = optarg;
 			break;
 
+		case 'r':
+			resolve_once = 1;
+			break;
+
 		case 'h':
 			hostname = optarg;
 			break;
@@ -133,6 +179,14 @@ int main(int argc, char *argv[])
 			timeout = atoi(optarg);
 			break;
 
+		case 'I':
+			useragent = optarg;
+			break;
+
+		case 'R':
+			referer = optarg;
+			break;
+
 		case 'f':
 			wait = 0;
 			break;
@@ -159,6 +213,22 @@ int main(int argc, char *argv[])
 			version();
 			return 0;
 
+		case 'n':
+		{
+			char *dummy = strchr(optarg, ',');
+			if (nagios_mode) error_exit("-n and -N are mutual exclusive\n");
+			nagios_mode = 1;
+			if (!dummy)
+				error_exit("-n: missing parameter\n");
+			nagios_warn = atof(optarg);
+			nagios_crit = atof(dummy + 1);
+		} break;
+		case 'N':
+			if (nagios_mode) error_exit("-n and -N are mutual exclusive\n");
+			nagios_mode = 2;
+			nagios_exit_code = atoi(optarg);
+			break;
+
 		case '?':
 		default:
 			usage();
@@ -166,6 +236,9 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	last_error = mymalloc(4096, "error buffer");
+	last_error[0] = 0x00;
+
 #ifndef NO_SSL
 	if (use_ssl && portnr == 80)
 		portnr = 443;
@@ -209,9 +282,8 @@ int main(int argc, char *argv[])
 
 	if (hostname == NULL)
 	{
-		fprintf(stderr, "No hostname/getrequest given\n");
 		usage();
-		return 3;
+		error_exit("No hostname/getrequest given\n");
 	}
 
 	if (get == NULL)
@@ -242,7 +314,8 @@ int main(int argc, char *argv[])
 			proxyport = atoi(dummy + 1);
 		}
 
-		fprintf(stderr, "Using proxyserver: %s:%d\n", proxyhost, proxyport);
+		if (!quiet && !nagios_mode)
+			fprintf(stderr, "Using proxyserver: %s:%d\n", proxyhost, proxyport);
 	}
 
 #ifndef NO_SSL
@@ -250,15 +323,25 @@ int main(int argc, char *argv[])
 	{
 		client_ctx = initialize_ctx();
 		if (!client_ctx)
-			error_exit("problem creating SSL context\n");
+		{
+			sprintf(last_error, "problem creating SSL context\n");
+			goto error_exit;
+		}
 	}
 #endif
 
 	request = mymalloc(strlen(get) + 4096, "request");
-	sprintf(request, "HEAD %s HTTP/1.0\r\nUser-Agent: HTTPing v" VERSION "\r\n\r\n", get);
+	sprintf(request, "HEAD %s HTTP/1.0\r\n", get);
+	if (useragent)
+		sprintf(&request[strlen(request)], "User-Agent: %s\r\n", useragent);
+	else
+		sprintf(&request[strlen(request)], "User-Agent: HTTPing v" VERSION "\r\n");
+	if (referer)
+		sprintf(&request[strlen(request)], "Referer: %s\r\n", referer);
+	strcat(request, "\r\n");
 	req_len = strlen(request);
 
-	if (!quiet && !machine_readable)
+	if (!quiet && !machine_readable && !nagios_mode)
 		printf("PING %s:%d (%s):\n", hostname, portnr, get);
 
 	signal(SIGINT, handler);
@@ -266,6 +349,21 @@ int main(int argc, char *argv[])
 
 	timeout *= 1000;	/* change to ms */
 
+	host = proxyhost?proxyhost:hostname;
+	port = proxyhost?proxyport:portnr;
+
+	if (resolve_once)
+	{
+		memset(&addr, 0x00, sizeof(addr));
+
+		if (resolve_host(host, &addr) == 0)
+		{
+			err++;
+			emit_error();
+			have_resolved = 1;
+		}
+	}
+
 	while((curncount < count || count == -1) && stop == 0)
 	{
 		double ms;
@@ -277,7 +375,9 @@ int main(int argc, char *argv[])
 
 		if (gettimeofday(&start, &tz) == -1)
 		{
-			perror("gettimeofday");
+			sprintf(last_error, "gettimeofday failed: %s\n", strerror(errno));
+			ok = 0;
+			emit_error();
 			break;
 		}
 
@@ -288,14 +388,30 @@ int main(int argc, char *argv[])
 			SSL *ssl_h = NULL;
 			BIO *s_bio;
 
-			if (proxyhost)
-				fd = connect_to(proxyhost, proxyport, timeout);
-			else
-				fd = connect_to(hostname, portnr, timeout);
+			curncount++;
+
+			if (!resolve_once || (resolve_once == 1 && have_resolved == 0))
+			{
+				memset(&addr, 0x00, sizeof(addr));
+
+				if (resolve_host(host, &addr) == -1)
+				{
+					err++;
+					emit_error();
+					break;
+				}
+
+				have_resolved = 1;
+			}
+
+			fd = connect_to(&addr, port, timeout);
 
 			if (fd == -3)	/* ^C pressed */
 				break;
 
+			if (fd < 0)
+				emit_error();
+
 			if (fd >= 0)
 			{
 				/* set fd blocking */
@@ -309,19 +425,20 @@ int main(int argc, char *argv[])
 
 					rc = connect_ssl(fd, client_ctx, &ssl_h, &s_bio, timeout);
 					if (rc != 0)
+					{
+						close(fd);
 						fd = rc;
+					}
 				}
 #endif
 			}
 
-			curncount++;
-
 			if (fd < 0)
 			{
-				if (fd == -1)
-					printf("error connecting to host\n");
-				else if (fd == -2)
-					printf("timeout connecting to host\n");
+				if (fd == -2)
+					sprintf(last_error, "timeout connecting to host\n");
+
+				emit_error();
 
 				err++;
 
@@ -337,13 +454,15 @@ int main(int argc, char *argv[])
 			if (rc != req_len)
 			{
 				if (rc == -1)
-					printf("error sending request to host\n");
+					sprintf(last_error, "error sending request to host\n");
 				else if (rc == -2)
-					printf("timeout sending to host\n");
+					sprintf(last_error, "timeout sending to host\n");
 				else if (rc == -3)
 				{/* ^C */}
 				else if (rc == 0)
-					printf("connection prematurely closed by peer\n");
+					sprintf(last_error, "connection prematurely closed by peer\n");
+
+				emit_error();
 
 				close(fd);
 				err++;
@@ -355,21 +474,24 @@ int main(int argc, char *argv[])
 
 			if (show_statuscodes || machine_readable)
 			{
-				/* statuscode is in first line behind
-				 * 'HTTP/1.x'
-				 */
-				char *dummy = strchr(reply, ' ');
-
-				if (dummy)
+				if (reply)
 				{
-					sc = strdup(dummy + 1);
-
-					/* lines are normally terminated with a
-					 * LF
+					/* statuscode is in first line behind
+					 * 'HTTP/1.x'
 					 */
-					dummy = strchr(sc, '\n');
+					char *dummy = strchr(reply, ' ');
+
 					if (dummy)
-						*dummy = 0x00;
+					{
+						sc = strdup(dummy + 1);
+
+						/* lines are normally terminated with a
+						 * LF
+						 */
+						dummy = strchr(sc, '\n');
+						if (dummy)
+							*dummy = 0x00;
+					}
 				}
 			}
 
@@ -378,9 +500,11 @@ int main(int argc, char *argv[])
 			if (rc < 0)
 			{
 				if (rc == RC_SHORTREAD)
-					printf("error receiving reply from host\n");
+					sprintf(last_error, "error receiving reply from host\n");
 				else if (rc == RC_TIMEOUT)
-					printf("timeout receiving reply from host\n");
+					sprintf(last_error, "timeout receiving reply from host\n");
+
+				emit_error();
 
 				close(fd);
 				err++;
@@ -392,7 +516,9 @@ int main(int argc, char *argv[])
 
 			if (gettimeofday(&end, &tz) == -1)
 			{
-				perror("gettimeofday");
+				sprintf(last_error, "gettimeofday failed: %s\n", strerror(errno));
+				ok = 0;
+				emit_error();
 				break;
 			}
 
@@ -400,7 +526,10 @@ int main(int argc, char *argv[])
 			if (use_ssl)
 			{
 				if (close_ssl_connection(ssl_h, fd) == -1)
-					error_exit("error shutting down ssl");
+				{
+					sprintf(last_error, "error shutting down ssl\n");
+					emit_error();
+				}
 			}
 #endif
 			close(fd);
@@ -414,21 +543,28 @@ int main(int argc, char *argv[])
 
 			if (machine_readable)
 			{
-				char *dummy = strchr(sc, ' ');
+				if (sc)
+				{
+					char *dummy = strchr(sc, ' ');
 
-				if (dummy) *dummy = 0x00;
+					if (dummy) *dummy = 0x00;
 
-				if (strstr(ok_str, sc))
-					printf("%f", ms);
+					if (strstr(ok_str, sc))
+						printf("%f", ms);
+					else
+						printf("%s", err_str);
+
+					if (show_statuscodes)
+						printf(" %s", sc);
+				}
 				else
+				{
 					printf("%s", err_str);
-
-				if (show_statuscodes)
-					printf(" %s", sc);
+				}
 
 				printf("\n");
 			}
-			else
+			else if (!quiet && !nagios_mode)
 			{
 				printf("connected to %s:%d, seq=%d time=%.2f ms %s\n", hostname, portnr, curncount-1, ms, sc?sc:"");
 			}
@@ -442,7 +578,9 @@ int main(int argc, char *argv[])
 			sleep(wait);
 	}
 
-	if (!quiet && !machine_readable)
+	avg_httping_time = avg / (double)ok;
+
+	if (!quiet && !machine_readable && !nagios_mode)
 	{
 		printf("--- %s ping statistics ---\n", get);
 		if (count == -1)
@@ -451,7 +589,40 @@ int main(int argc, char *argv[])
 			printf("%d connects, %d ok, %3.2f%% failed\n", curncount, ok, (((double)err) / ((double)count)) * 100.0);
 
 		if (ok > 0)
-			printf("round-trip min/avg/max = %.1f/%.1f/%.1f ms\n", min, avg / (double)ok, max);
+			printf("round-trip min/avg/max = %.1f/%.1f/%.1f ms\n", min, avg_httping_time, max);
+	}
+
+error_exit:
+	if (nagios_mode == 1)
+	{
+		if (ok == 0)
+		{
+			printf("CRITICAL - connecting failed: %s", last_error);
+			return 2;
+		}
+		else if (avg_httping_time >= nagios_crit)
+		{
+			printf("CRITICAL - average httping-time is %.1f\n", avg_httping_time);
+			return 2;
+		}
+		else if (avg_httping_time >= nagios_warn)
+		{
+			printf("WARNING - average httping-time is %.1f\n", avg_httping_time);
+			return 1;
+		}
+
+		printf("OK - average httping-time is %.1f (%s)\n", avg_httping_time, last_error);
+	}
+	else if (nagios_mode == 2)
+	{
+		if (ok && last_error[0] == 0x00)
+		{
+			printf("OK - all fine, avg httping time is %.1f\n", avg_httping_time);
+			return 0;
+		}
+
+		printf("%s: - failed: %s", nagios_exit_code == 1?"WARNING":(nagios_exit_code == 2?"CRITICAL":"ERROR"), last_error);
+		return nagios_exit_code;
 	}
 
 	if (ok)
