@@ -13,6 +13,4 @@
   files in the program, then also delete it here.
 */
 
-#define incopy(a)       *((struct in_addr *)a)
-
-int connect_to(char *host, int portnr, int timeout);
+int connect_to(struct sockaddr_in *addr, int portnr, int timeout);
